@@ -1,14 +1,12 @@
 import warnings
 
-from UserList import UserList
+from kombu.utils import cached_property
 
-from celery import conf
 from celery import registry
+from celery.app import app_or_default
 from celery.datastructures import AttributeDict
-from celery.messaging import with_connection
-from celery.messaging import TaskPublisher
-from celery.result import TaskSetResult
 from celery.utils import gen_unique_id
+from celery.utils.compat import UserList
 
 TASKSET_DEPRECATION_TEXT = """\
 Using this invocation of TaskSet is deprecated and will be removed
@@ -22,7 +20,6 @@ this so the syntax has been changed to:
     ts = TaskSet(tasks=[
             %(cls)s.subtask(args1, kwargs1, options1),
             %(cls)s.subtask(args2, kwargs2, options2),
-            %(cls)s.subtask(args3, kwargs3, options3),
             ...
             %(cls)s.subtask(argsN, kwargsN, optionsN),
     ])
@@ -57,13 +54,11 @@ class subtask(AttributeDict):
 
     """
 
-    def __init__(self, task=None, args=None, kwargs=None, options=None,
-            **extra):
+    def __init__(self, task=None, args=None, kwargs=None, options=None, **ex):
         init = super(subtask, self).__init__
 
         if isinstance(task, dict):
-            # Use the values from a dict.
-            return init(task)
+            return init(task)  # works like dict(d)
 
         # Also supports using task class/instance instead of string name.
         try:
@@ -72,11 +67,11 @@ class subtask(AttributeDict):
             task_name = task
 
         init(task=task_name, args=tuple(args or ()),
-                             kwargs=dict(kwargs or {}, **extra),
+                             kwargs=dict(kwargs or {}, **ex),
                              options=options or {})
 
     def delay(self, *argmerge, **kwmerge):
-        """Shortcut to ``apply_async(argmerge, kwargs)``."""
+        """Shortcut to `apply_async(argmerge, kwargs)`."""
         return self.apply_async(args=argmerge, kwargs=kwmerge)
 
     def apply(self, args=(), kwargs={}, **options):
@@ -85,7 +80,7 @@ class subtask(AttributeDict):
         args = tuple(args) + tuple(self.args)
         kwargs = dict(self.kwargs, **kwargs)
         options = dict(self.options, **options)
-        return self.get_type().apply(args, kwargs, **options)
+        return self.type.apply(args, kwargs, **options)
 
     def apply_async(self, args=(), kwargs={}, **options):
         """Apply this task asynchronously."""
@@ -93,44 +88,48 @@ class subtask(AttributeDict):
         args = tuple(args) + tuple(self.args)
         kwargs = dict(self.kwargs, **kwargs)
         options = dict(self.options, **options)
-        return self.get_type().apply_async(args, kwargs, **options)
+        return self.type.apply_async(args, kwargs, **options)
 
     def get_type(self):
-        # For JSON serialization, the task class is lazily loaded,
+        return self.type
+
+    def __reduce__(self):
+        # for serialization, the task type is lazily loaded,
         # and not stored in the dict itself.
+        return (self.__class__, (dict(self), ), None)
+
+    def __repr__(self, kwformat=lambda i: "%s=%r" % i, sep=', '):
+        kw = self["kwargs"]
+        return "%s(%s%s%s)" % (self["task"], sep.join(map(repr, self["args"])),
+                kw and sep or "", sep.join(map(kwformat, kw.iteritems())))
+
+    @cached_property
+    def type(self):
         return registry.tasks[self.task]
 
 
 class TaskSet(UserList):
     """A task containing several subtasks, making it possible
-    to track how many, or when all of the tasks has been completed.
+    to track how many, or when all of the tasks have been completed.
 
     :param tasks: A list of :class:`subtask` instances.
 
-    .. attribute:: total
-
-        Total number of subtasks in this task set.
-
     Example::
 
-        >>> from djangofeeds.tasks import RefreshFeedTask
-        >>> from celery.task.sets import TaskSet, subtask
-        >>> urls = ("http://cnn.com/rss",
-        ...         "http://bbc.co.uk/rss",
-        ...         "http://xkcd.com/rss")
-        >>> subtasks = [RefreshFeedTask.subtask(kwargs={"feed_url": url})
-        ...                 for url in urls]
-        >>> taskset = TaskSet(tasks=subtasks)
+        >>> urls = ("http://cnn.com/rss", "http://bbc.co.uk/rss")
+        >>> taskset = TaskSet(refresh_feed.subtask((url, )) for url in urls)
         >>> taskset_result = taskset.apply_async()
-        >>> list_of_return_values = taskset_result.join()
+        >>> list_of_return_values = taskset_result.join()  # *expensive*
 
     """
-    Publisher = TaskPublisher
+    _task = None                # compat
+    _task_name = None           # compat
 
-    _task = None                                                # compat
-    _task_name = None                                           # compat
+    #: Total number of subtasks in this set.
+    total = None
 
-    def __init__(self, task=None, tasks=None):
+    def __init__(self, task=None, tasks=None, app=None, Publisher=None):
+        self.app = app_or_default(app)
         if task is not None:
             if hasattr(task, "__iter__"):
                 tasks = task
@@ -144,63 +143,40 @@ class TaskSet(UserList):
                 warnings.warn(TASKSET_DEPRECATION_TEXT % {
                                 "cls": task.__class__.__name__},
                               DeprecationWarning)
-
-        self.data = list(tasks)
+        self.data = list(tasks or [])
         self.total = len(self.tasks)
+        self.Publisher = Publisher or self.app.amqp.TaskPublisher
 
-    @with_connection
-    def apply_async(self, connection=None,
-            connect_timeout=conf.BROKER_CONNECTION_TIMEOUT):
-        """Run all tasks in the taskset.
+    def apply_async(self, connection=None, connect_timeout=None,
+            publisher=None, taskset_id=None):
+        """Apply taskset."""
+        return self.app.with_default_connection(self._apply_async)(
+                    connection=connection,
+                    connect_timeout=connect_timeout,
+                    publisher=publisher,
+                    taskset_id=taskset_id)
 
-        Returns a :class:`celery.result.TaskSetResult` instance.
+    def _apply_async(self, connection=None, connect_timeout=None,
+            publisher=None, taskset_id=None):
+        if self.app.conf.CELERY_ALWAYS_EAGER:
+            return self.apply(taskset_id=taskset_id)
 
-        Example
-
-            >>> ts = TaskSet(tasks=(
-            ...         RefreshFeedTask.subtask(["http://foo.com/rss"]),
-            ...         RefreshFeedTask.subtask(["http://bar.com/rss"]),
-            ... ))
-            >>> result = ts.apply_async()
-            >>> result.taskset_id
-            "d2c9b261-8eff-4bfb-8459-1e1b72063514"
-            >>> result.subtask_ids
-            ["b4996460-d959-49c8-aeb9-39c530dcde25",
-            "598d2d18-ab86-45ca-8b4f-0779f5d6a3cb"]
-            >>> result.waiting()
-            True
-            >>> time.sleep(10)
-            >>> result.ready()
-            True
-            >>> result.successful()
-            True
-            >>> result.failed()
-            False
-            >>> result.join()
-            [True, True]
-
-        """
-        if conf.ALWAYS_EAGER:
-            return self.apply()
-
-        taskset_id = gen_unique_id()
-        publisher = self.Publisher(connection=connection)
+        setid = taskset_id or gen_unique_id()
+        pub = publisher or self.Publisher(connection=connection)
         try:
-            results = [task.apply_async(taskset_id=taskset_id,
-                                        publisher=publisher)
+            results = [task.apply_async(taskset_id=setid, publisher=pub)
                             for task in self.tasks]
         finally:
-            publisher.close()
+            if not publisher:  # created by us.
+                pub.close()
 
-        return TaskSetResult(taskset_id, results)
+        return self.app.TaskSetResult(setid, results)
 
-    def apply(self):
-        """Applies the taskset locally."""
-        taskset_id = gen_unique_id()
-
-        # This will be filled with EagerResults.
-        return TaskSetResult(taskset_id, [task.apply(taskset_id=taskset_id)
-                                            for task in self.tasks])
+    def apply(self, taskset_id=None):
+        """Applies the taskset locally by blocking until all tasks return."""
+        setid = taskset_id or gen_unique_id()
+        return self.app.TaskSetResult(setid, [task.apply(taskset_id=setid)
+                                                for task in self.tasks])
 
     @property
     def tasks(self):
