@@ -1,28 +1,38 @@
 import sys
+
 from datetime import datetime
 
-from celery import conf
-from celery import log
+from celery.platforms import get_signal
 from celery.registry import tasks
 from celery.utils import timeutils
 from celery.worker import state
 from celery.worker.state import revoked
 from celery.worker.control.registry import Panel
+from celery.utils.encoding import safe_repr
 
 TASK_INFO_FIELDS = ("exchange", "routing_key", "rate_limit")
 
 
 @Panel.register
-def revoke(panel, task_id, **kwargs):
+def revoke(panel, task_id, terminate=False, signal=None, **kwargs):
     """Revoke task by task id."""
     revoked.add(task_id)
-    panel.logger.warn("Task %s revoked" % (task_id, ))
-    return {"ok": "task %s revoked" % (task_id, )}
+    action = "revoked"
+    if terminate:
+        signum = get_signal(signal)
+        for request in state.active_requests:
+            if request.task_id == task_id:
+                action = "terminated (%s)" % (signum, )
+                request.terminate(panel.consumer.pool, signal=signum)
+                break
+
+    panel.logger.warn("Task %s %s." % (task_id, action))
+    return {"ok": "task %s %s" % (task_id, action)}
 
 
 @Panel.register
 def enable_events(panel):
-    dispatcher = panel.listener.event_dispatcher
+    dispatcher = panel.consumer.event_dispatcher
     if not dispatcher.enabled:
         dispatcher.enable()
         dispatcher.send("worker-online")
@@ -33,7 +43,7 @@ def enable_events(panel):
 
 @Panel.register
 def disable_events(panel):
-    dispatcher = panel.listener.event_dispatcher
+    dispatcher = panel.consumer.event_dispatcher
     if dispatcher.enabled:
         dispatcher.send("worker-offline")
         dispatcher.disable()
@@ -44,19 +54,9 @@ def disable_events(panel):
 
 @Panel.register
 def heartbeat(panel):
-    panel.logger.info("Heartbeat requested by remote.")
-    dispatcher = panel.listener.event_dispatcher
-    if dispatcher.enabled:
-        dispatcher.send("worker-heartbeat")
-
-
-@Panel.register
-def set_loglevel(panel, loglevel=None):
-    if loglevel is not None:
-        if not isinstance(loglevel, int):
-            loglevel = conf.LOG_LEVELS[loglevel.upper()]
-        log.get_default_logger(loglevel=loglevel)
-    return {"ok": loglevel}
+    panel.logger.debug("Heartbeat requested by remote.")
+    dispatcher = panel.consumer.event_dispatcher
+    dispatcher.send("worker-heartbeat", **state.SOFTWARE_INFO)
 
 
 @Panel.register
@@ -82,11 +82,11 @@ def rate_limit(panel, task_name, rate_limit, **kwargs):
             task_name, ), exc_info=sys.exc_info())
         return {"error": "unknown task"}
 
-    if conf.DISABLE_RATE_LIMITS:
+    if not hasattr(panel.consumer.ready_queue, "refresh"):
         panel.logger.error("Rate limit attempt, but rate limits disabled.")
         return {"error": "rate limits disabled"}
 
-    panel.listener.ready_queue.refresh()
+    panel.consumer.ready_queue.refresh()
 
     if not rate_limit:
         panel.logger.warn("Disabled rate limits for tasks of type %s" % (
@@ -100,7 +100,7 @@ def rate_limit(panel, task_name, rate_limit, **kwargs):
 
 @Panel.register
 def dump_schedule(panel, safe=False, **kwargs):
-    schedule = panel.listener.eta_schedule.schedule
+    schedule = panel.consumer.eta_schedule.schedule
     if not schedule.queue:
         panel.logger.info("--Empty schedule--")
         return []
@@ -123,13 +123,13 @@ def dump_schedule(panel, safe=False, **kwargs):
 
 @Panel.register
 def dump_reserved(panel, safe=False, **kwargs):
-    ready_queue = panel.listener.ready_queue
+    ready_queue = panel.consumer.ready_queue
     reserved = ready_queue.items
     if not reserved:
         panel.logger.info("--Empty queue--")
         return []
     panel.logger.info("* Dump of currently reserved tasks:\n%s" % (
-                            "\n".join(map(repr, reserved), )))
+                            "\n".join(map(safe_repr, reserved), )))
     return [request.info(safe=safe)
             for request in reserved]
 
@@ -143,8 +143,8 @@ def dump_active(panel, safe=False, **kwargs):
 @Panel.register
 def stats(panel, **kwargs):
     return {"total": state.total_count,
-            "listener": panel.listener.info,
-            "pool": panel.listener.pool.info}
+            "consumer": panel.consumer.info,
+            "pool": panel.consumer.pool.info}
 
 
 @Panel.register
@@ -177,6 +177,18 @@ def ping(panel, **kwargs):
     return "pong"
 
 
+@Panel.register
+def pool_grow(panel, n=1, **kwargs):
+    panel.consumer.pool.grow(n)
+    return {"ok": "spawned worker processes"}
+
+
+@Panel.register
+def pool_shrink(panel, n=1, **kwargs):
+    panel.consumer.pool.shrink(n)
+    return {"ok": "terminated worker processes"}
+
+
 @Panel.register
 def shutdown(panel, **kwargs):
     panel.logger.critical("Got shutdown from remote.")
@@ -186,7 +198,7 @@ def shutdown(panel, **kwargs):
 @Panel.register
 def add_consumer(panel, queue=None, exchange=None, exchange_type="direct",
         routing_key=None, **options):
-    cset = panel.listener.task_consumer
+    cset = panel.consumer.task_consumer
     declaration = dict(queue=queue,
                        exchange=exchange,
                        exchange_type=exchange_type,
@@ -200,6 +212,13 @@ def add_consumer(panel, queue=None, exchange=None, exchange_type="direct",
 
 @Panel.register
 def cancel_consumer(panel, queue=None, **_):
-    cset = panel.listener.task_consumer
+    cset = panel.consumer.task_consumer
     cset.cancel_by_queue(queue)
     return {"ok": "no longer consuming from %s" % (queue, )}
+
+
+@Panel.register
+def active_queues(panel):
+    """Returns the queues associated with each worker."""
+    return [dict(queue.as_dict(recurse=True))
+                    for queue in panel.consumer.task_consumer.queues]
