@@ -7,22 +7,25 @@ import warnings
 
 from datetime import datetime
 
-from celery import log
 from celery import platforms
+from celery.app import app_or_default, current_app
 from celery.datastructures import ExceptionInfo
 from celery.exceptions import SoftTimeLimitExceeded, TimeLimitExceeded
 from celery.exceptions import WorkerLostError, RetryTaskError
 from celery.execute.trace import TaskTrace
-from celery.loaders import current_loader
 from celery.registry import tasks
 from celery.utils import noop, kwdict, fun_takes_kwargs
-from celery.utils import truncate_text, maybe_iso8601
+from celery.utils import truncate_text
 from celery.utils.compat import log_with_extra
+from celery.utils.encoding import safe_repr, safe_str
+from celery.utils.timeutils import maybe_iso8601
 from celery.worker import state
 
 # pep8.py borks on a inline signature separator and
 # says "trailing whitespace" ;)
 EMAIL_SIGNATURE_SEP = "-- "
+
+#: format string for the body of an error e-mail.
 TASK_ERROR_EMAIL_BODY = """
 Task %%(name)s with id %%(id)s raised exception:\n%%(exc)r
 
@@ -39,6 +42,8 @@ celeryd at %%(hostname)s.
 """ % {"EMAIL_SIGNATURE_SEP": EMAIL_SIGNATURE_SEP}
 
 
+#: Keys to keep from the message delivery info.  The values
+#: of these keys must be pickleable.
 WANTED_DELIVERY_INFO = ("exchange", "routing_key", "consumer_tag", )
 
 
@@ -57,27 +62,38 @@ class WorkerTaskTrace(TaskTrace):
     meta backend.
 
     If the call was successful, it saves the result to the task result
-    backend, and sets the task status to ``"SUCCESS"``.
+    backend, and sets the task status to `"SUCCESS"`.
 
     If the call raises :exc:`celery.exceptions.RetryTaskError`, it extracts
     the original exception, uses that as the result and sets the task status
-    to ``"RETRY"``.
+    to `"RETRY"`.
 
     If the call results in an exception, it saves the exception as the task
-    result, and sets the task status to ``"FAILURE"``.
+    result, and sets the task status to `"FAILURE"`.
 
     :param task_name: The name of the task to execute.
     :param task_id: The unique id of the task.
     :param args: List of positional args to pass on to the function.
     :param kwargs: Keyword arguments mapping to pass on to the function.
 
+    :keyword loader: Custom loader to use, if not specified the current app
+      loader will be used.
+    :keyword hostname: Custom hostname to use, if not specified the system
+      hostname will be used.
+
     :returns: the evaluated functions return value on success, or
         the exception instance on failure.
 
     """
 
+    #: Current loader.
+    loader = None
+
+    #: Hostname to report as.
+    hostname = None
+
     def __init__(self, *args, **kwargs):
-        self.loader = kwargs.get("loader") or current_loader()
+        self.loader = kwargs.get("loader") or current_app().loader
         self.hostname = kwargs.get("hostname") or socket.gethostname()
         super(WorkerTaskTrace, self).__init__(*args, **kwargs)
 
@@ -143,90 +159,98 @@ def execute_and_trace(task_name, *args, **kwargs):
 
     """
     hostname = kwargs.get("hostname")
-    platforms.set_mp_process_title("celeryd", info=task_name,
-                                   hostname=hostname)
+    platforms.set_mp_process_title("celeryd", task_name, hostname=hostname)
     try:
         return WorkerTaskTrace(task_name, *args, **kwargs).execute_safe()
     finally:
-        platforms.set_mp_process_title("celeryd", hostname=hostname)
+        platforms.set_mp_process_title("celeryd", "-idle-", hostname)
 
 
 class TaskRequest(object):
-    """A request for task execution.
+    """A request for task execution."""
 
-    :param task_name: see :attr:`task_name`.
-    :param task_id: see :attr:`task_id`.
-    :param args: see :attr:`args`
-    :param kwargs: see :attr:`kwargs`.
+    #: Kind of task.  Must be a name registered in the task registry.
+    name = None
 
-    .. attribute:: task_name
+    #: The task class (set by constructor using :attr:`task_name`).
+    task = None
 
-        Kind of task. Must be a name registered in the task registry.
+    #: UUID of the task.
+    task_id = None
 
-    .. attribute:: task_id
+    #: UUID of the taskset that this task belongs to.
+    taskset_id = None
 
-        UUID of the task.
+    #: List of positional arguments to apply to the task.
+    args = None
 
-    .. attribute:: args
+    #: Mapping of keyword arguments to apply to the task.
+    kwargs = None
 
-        List of positional arguments to apply to the task.
+    #: Number of times the task has been retried.
+    retries = 0
 
-    .. attribute:: kwargs
+    #: The tasks eta (for information only).
+    eta = None
 
-        Mapping of keyword arguments to apply to the task.
+    #: When the task expires.
+    expires = None
 
-    .. attribute:: on_ack
+    #: Callback called when the task should be acknowledged.
+    on_ack = None
 
-        Callback called when the task should be acknowledged.
+    #: The message object.  Used to acknowledge the message.
+    message = None
 
-    .. attribute:: message
+    #: Flag set when the task has been executed.
+    executed = False
 
-        The original message sent. Used for acknowledging the message.
+    #: Additional delivery info, e.g. contains the path from
+    #: Producer to consumer.
+    delivery_info = None
 
-    .. attribute:: executed
+    #: Flag set when the task has been acknowledged.
+    acknowledged = False
 
-        Set to :const:`True` if the task has been executed.
-        A task should only be executed once.
-
-    .. attribute:: delivery_info
-
-        Additional delivery info, e.g. the contains the path
-        from producer to consumer.
-
-    .. attribute:: acknowledged
-
-        Set to :const:`True` if the task has been acknowledged.
-
-    """
-    # Logging output
+    #: Format string used to log task success.
     success_msg = """\
         Task %(name)s[%(id)s] succeeded in %(runtime)ss: %(return_value)s
     """
+
+    #: Format string used to log task failure.
     error_msg = """\
         Task %(name)s[%(id)s] raised exception: %(exc)s\n%(traceback)s
     """
-    retry_msg = """\
-        Task %(name)s[%(id)s] retry: %(exc)s
-    """
 
-    # E-mails
+    #: Format string used to log task retry.
+    retry_msg = """Task %(name)s[%(id)s] retry: %(exc)s"""
+
+    #: Format string used to generate error e-mail subjects.
     email_subject = """\
         [celery@%(hostname)s] Error: Task %(name)s (%(id)s): %(exc)s
     """
+
+    #: Format string used to generate error e-mail content.
     email_body = TASK_ERROR_EMAIL_BODY
 
-    # Internal flags
-    executed = False
-    acknowledged = False
+    #: Timestamp set when the task is started.
     time_start = None
+
+    #: Process id of the worker processing this task (if any).
+    worker_pid = None
+
     _already_revoked = False
+    _terminate_on_ack = None
 
     def __init__(self, task_name, task_id, args, kwargs,
             on_ack=noop, retries=0, delivery_info=None, hostname=None,
             email_subject=None, email_body=None, logger=None,
-            eventer=None, eta=None, expires=None, **opts):
+            eventer=None, eta=None, expires=None, app=None,
+            taskset_id=None, **opts):
+        self.app = app_or_default(app)
         self.task_name = task_name
         self.task_id = task_id
+        self.taskset_id = taskset_id
         self.retries = retries
         self.args = args
         self.kwargs = kwargs
@@ -235,7 +259,7 @@ class TaskRequest(object):
         self.on_ack = on_ack
         self.delivery_info = delivery_info or {}
         self.hostname = hostname or socket.gethostname()
-        self.logger = logger or log.get_default_logger()
+        self.logger = logger or self.app.log.get_default_logger()
         self.eventer = eventer
         self.email_subject = email_subject or self.email_subject
         self.email_body = email_body or self.email_body
@@ -245,68 +269,54 @@ class TaskRequest(object):
         if self.task.ignore_result:
             self._store_errors = self.task.store_errors_even_if_ignored
 
-    def maybe_expire(self):
-        if self.expires and datetime.now() > self.expires:
-            state.revoked.add(self.task_id)
-            if self._store_errors:
-                self.task.backend.mark_as_revoked(self.task_id)
-
-    def revoked(self):
-        if self._already_revoked:
-            return True
-        if self.expires:
-            self.maybe_expire()
-        if self.task_id in state.revoked:
-            self.logger.warn("Skipping revoked task: %s[%s]" % (
-                self.task_name, self.task_id))
-            self.send_event("task-revoked", uuid=self.task_id)
-            self.acknowledge()
-            self._already_revoked = True
-            return True
-        return False
-
     @classmethod
-    def from_message(cls, message, message_data, on_ack=noop, logger=None,
-            eventer=None, hostname=None):
-        """Create a :class:`TaskRequest` from a task message sent by
-        :class:`celery.messaging.TaskPublisher`.
+    def from_message(cls, message, body, on_ack=noop, **kw):
+        """Create request from a task message.
 
         :raises UnknownTaskError: if the message does not describe a task,
             the message is also rejected.
 
-        :returns :class:`TaskRequest`:
-
         """
-        task_name = message_data["task"]
-        task_id = message_data["id"]
-        args = message_data["args"]
-        kwargs = message_data["kwargs"]
-        retries = message_data.get("retries", 0)
-        eta = maybe_iso8601(message_data.get("eta"))
-        expires = maybe_iso8601(message_data.get("expires"))
-
         _delivery_info = getattr(message, "delivery_info", {})
         delivery_info = dict((key, _delivery_info.get(key))
                                 for key in WANTED_DELIVERY_INFO)
 
+        kwargs = body["kwargs"]
         if not hasattr(kwargs, "items"):
-            raise InvalidTaskError("Task kwargs must be a dictionary.")
+            raise InvalidTaskError("Task keyword arguments is not a mapping.")
 
-        return cls(task_name, task_id, args, kwdict(kwargs),
-                   retries=retries, on_ack=on_ack,
-                   delivery_info=delivery_info, logger=logger,
-                   eventer=eventer, hostname=hostname,
-                   eta=eta, expires=expires)
+        return cls(task_name=body["task"],
+                   task_id=body["id"],
+                   taskset_id=body.get("taskset", None),
+                   args=body["args"],
+                   kwargs=kwdict(kwargs),
+                   retries=body.get("retries", 0),
+                   eta=maybe_iso8601(body.get("eta")),
+                   expires=maybe_iso8601(body.get("expires")),
+                   on_ack=on_ack,
+                   delivery_info=delivery_info,
+                   **kw)
+
+    def get_instance_attrs(self, loglevel, logfile):
+        return {"logfile": logfile,
+                "loglevel": loglevel,
+                "id": self.task_id,
+                "taskset": self.taskset_id,
+                "retries": self.retries,
+                "is_eager": False,
+                "delivery_info": self.delivery_info}
 
     def extend_with_default_kwargs(self, loglevel, logfile):
         """Extend the tasks keyword arguments with standard task arguments.
 
-        Currently these are ``logfile``, ``loglevel``, ``task_id``,
-        ``task_name``, ``task_retries``, and ``delivery_info``.
+        Currently these are `logfile`, `loglevel`, `task_id`,
+        `task_name`, `task_retries`, and `delivery_info`.
 
         See :meth:`celery.task.base.Task.run` for more information.
 
         """
+        if not self.task.accept_magic_kwargs:
+            return self.kwargs
         kwargs = dict(self.kwargs)
         default_kwargs = {"logfile": logfile,
                           "loglevel": loglevel,
@@ -322,46 +332,6 @@ class TaskRequest(object):
         kwargs.update(extend_with)
         return kwargs
 
-    def _get_tracer_args(self, loglevel=None, logfile=None):
-        """Get the :class:`WorkerTaskTrace` tracer for this task."""
-        task_func_kwargs = self.extend_with_default_kwargs(loglevel, logfile)
-        return self.task_name, self.task_id, self.args, task_func_kwargs
-
-    def _set_executed_bit(self):
-        """Set task as executed to make sure it's not executed again."""
-        if self.executed:
-            raise AlreadyExecutedError(
-                   "Task %s[%s] has already been executed" % (
-                       self.task_name, self.task_id))
-        self.executed = True
-
-    def execute(self, loglevel=None, logfile=None):
-        """Execute the task in a :class:`WorkerTaskTrace`.
-
-        :keyword loglevel: The loglevel used by the task.
-
-        :keyword logfile: The logfile used by the task.
-
-        """
-        if self.revoked():
-            return
-        # Make sure task has not already been executed.
-        self._set_executed_bit()
-
-        # acknowledge task as being processed.
-        if not self.task.acks_late:
-            self.acknowledge()
-
-        tracer = WorkerTaskTrace(*self._get_tracer_args(loglevel, logfile),
-                                 **{"hostname": self.hostname})
-        retval = tracer.execute()
-        self.acknowledge()
-        return retval
-
-    def send_event(self, type, **fields):
-        if self.eventer:
-            self.eventer.send(type, **fields)
-
     def execute_using_pool(self, pool, loglevel=None, logfile=None):
         """Like :meth:`execute`, but using the :mod:`multiprocessing` pool.
 
@@ -378,26 +348,94 @@ class TaskRequest(object):
         self._set_executed_bit()
 
         args = self._get_tracer_args(loglevel, logfile)
+        instance_attrs = self.get_instance_attrs(loglevel, logfile)
         result = pool.apply_async(execute_and_trace,
                                   args=args,
-                                  kwargs={"hostname": self.hostname},
+                                  kwargs={"hostname": self.hostname,
+                                          "request": instance_attrs},
                                   accept_callback=self.on_accepted,
                                   timeout_callback=self.on_timeout,
                                   callbacks=[self.on_success],
                                   errbacks=[self.on_failure])
         return result
 
-    def on_accepted(self):
+    def execute(self, loglevel=None, logfile=None):
+        """Execute the task in a :class:`WorkerTaskTrace`.
+
+        :keyword loglevel: The loglevel used by the task.
+
+        :keyword logfile: The logfile used by the task.
+
+        """
+        if self.revoked():
+            return
+
+        # Make sure task has not already been executed.
+        self._set_executed_bit()
+
+        # acknowledge task as being processed.
+        if not self.task.acks_late:
+            self.acknowledge()
+
+        instance_attrs = self.get_instance_attrs(loglevel, logfile)
+        tracer = WorkerTaskTrace(*self._get_tracer_args(loglevel, logfile),
+                                 **{"hostname": self.hostname,
+                                    "loader": self.app.loader,
+                                    "request": instance_attrs})
+        retval = tracer.execute()
+        self.acknowledge()
+        return retval
+
+    def maybe_expire(self):
+        """If expired, mark the task as revoked."""
+        if self.expires and datetime.now() > self.expires:
+            state.revoked.add(self.task_id)
+            if self._store_errors:
+                self.task.backend.mark_as_revoked(self.task_id)
+
+    def terminate(self, pool, signal=None):
+        if self._terminate_on_ack is not None:
+            return
+        elif self.time_start:
+            return pool.terminate_job(self.worker_pid, signal)
+        else:
+            self._terminate_on_ack = (True, pool, signal)
+
+    def revoked(self):
+        """If revoked, skip task and mark state."""
+        if self._already_revoked:
+            return True
+        if self.expires:
+            self.maybe_expire()
+        if self.task_id in state.revoked:
+            self.logger.warn("Skipping revoked task: %s[%s]" % (
+                self.task_name, self.task_id))
+            self.send_event("task-revoked", uuid=self.task_id)
+            self.acknowledge()
+            self._already_revoked = True
+            return True
+        return False
+
+    def send_event(self, type, **fields):
+        if self.eventer:
+            self.eventer.send(type, **fields)
+
+    def on_accepted(self, pid, time_accepted):
         """Handler called when task is accepted by worker pool."""
-        self.time_start = time.time()
+        self.worker_pid = pid
+        self.time_start = time_accepted
         state.task_accepted(self)
         if not self.task.acks_late:
             self.acknowledge()
-        self.send_event("task-started", uuid=self.task_id)
-        self.logger.debug("Task accepted: %s[%s]" % (
-            self.task_name, self.task_id))
+        self.send_event("task-started", uuid=self.task_id, pid=pid)
+        self.logger.debug("Task accepted: %s[%s] pid:%r" % (
+            self.task_name, self.task_id, pid))
+        if self._terminate_on_ack is not None:
+            _, pool, signal = self._terminate_on_ack
+            self.terminate(pool, signal)
 
     def on_timeout(self, soft):
+        """Handler called if the task times out."""
         state.task_ready(self)
         if soft:
             self.logger.warning("Soft time limit exceeded for %s[%s]" % (
@@ -411,47 +449,36 @@ class TaskRequest(object):
         if self._store_errors:
             self.task.backend.mark_as_failure(self.task_id, exc)
 
-    def acknowledge(self):
-        if not self.acknowledged:
-            self.on_ack()
-            self.acknowledged = True
-
     def on_success(self, ret_value):
-        """The handler used if the task was successfully processed (
-        without raising an exception)."""
+        """Handler called if the task was successfully processed."""
         state.task_ready(self)
 
         if self.task.acks_late:
             self.acknowledge()
 
-        runtime = time.time() - self.time_start
+        runtime = self.time_start and (time.time() - self.time_start) or 0
         self.send_event("task-succeeded", uuid=self.task_id,
-                        result=repr(ret_value), runtime=runtime)
+                        result=safe_repr(ret_value), runtime=runtime)
 
-        msg = self.success_msg.strip() % {
+        self.logger.info(self.success_msg.strip() % {
                 "id": self.task_id,
                 "name": self.task_name,
                 "return_value": self.repr_result(ret_value),
-                "runtime": runtime}
-        self.logger.info(msg)
-
-    def repr_result(self, result, maxlen=46):
-        # 46 is the length needed to fit
-        #     "the quick brown fox jumps over the lazy dog" :)
-        return truncate_text(repr(result), maxlen)
+                "runtime": runtime})
 
     def on_retry(self, exc_info):
+        """Handler called if the task should be retried."""
         self.send_event("task-retried", uuid=self.task_id,
-                                        exception=repr(exc_info.exception.exc),
-                                        traceback=repr(exc_info.traceback))
-        msg = self.retry_msg.strip() % {
+                         exception=safe_repr(exc_info.exception.exc),
+                         traceback=safe_repr(exc_info.traceback))
+
+        self.logger.info(self.retry_msg.strip() % {
                 "id": self.task_id,
                 "name": self.task_name,
-                "exc": repr(exc_info.exception.exc)}
-        self.logger.info(msg)
+                "exc": safe_repr(exc_info.exception.exc)})
 
     def on_failure(self, exc_info):
-        """The handler used if the task raised an exception."""
+        """Handler called if the task raised an exception."""
         state.task_ready(self)
 
         if self.task.acks_late:
@@ -468,14 +495,14 @@ class TaskRequest(object):
                                                   exc_info.exception)
 
         self.send_event("task-failed", uuid=self.task_id,
-                                       exception=repr(exc_info.exception),
-                                       traceback=exc_info.traceback)
+                         exception=safe_repr(exc_info.exception),
+                         traceback=safe_str(exc_info.traceback))
 
         context = {"hostname": self.hostname,
                    "id": self.task_id,
                    "name": self.task_name,
-                   "exc": repr(exc_info.exception),
-                   "traceback": unicode(exc_info.traceback, 'utf-8'),
+                   "exc": safe_repr(exc_info.exception),
+                   "traceback": safe_str(exc_info.traceback),
                    "args": self.args,
                    "kwargs": self.kwargs}
 
@@ -491,6 +518,12 @@ class TaskRequest(object):
                               enabled=task_obj.send_error_emails,
                               whitelist=task_obj.error_whitelist)
 
+    def acknowledge(self):
+        """Acknowledge task."""
+        if not self.acknowledged:
+            self.on_ack()
+            self.acknowledged = True
+
     def send_error_email(self, task, context, exc,
             whitelist=None, enabled=False, fail_silently=True):
         if enabled and not task.disable_error_emails:
@@ -499,21 +532,19 @@ class TaskRequest(object):
                     return
             subject = self.email_subject.strip() % context
             body = self.email_body.strip() % context
-            return current_loader().mail_admins(subject, body,
-                                                fail_silently=fail_silently)
+            self.app.mail_admins(subject, body, fail_silently=fail_silently)
 
-    def __repr__(self):
-        return '<%s: {name:"%s", id:"%s", args:"%s", kwargs:"%s"}>' % (
-                self.__class__.__name__,
-                self.task_name, self.task_id,
-                self.args, self.kwargs)
+    def repr_result(self, result, maxlen=46):
+        # 46 is the length needed to fit
+        #     "the quick brown fox jumps over the lazy dog" :)
+        return truncate_text(safe_repr(result), maxlen)
 
     def info(self, safe=False):
         args = self.args
         kwargs = self.kwargs
         if not safe:
-            args = repr(args)
-            kwargs = repr(self.kwargs)
+            args = safe_repr(args)
+            kwargs = safe_repr(self.kwargs)
 
         return {"id": self.task_id,
                 "name": self.task_name,
@@ -530,3 +561,21 @@ class TaskRequest(object):
                     self.task_id,
                     self.eta and " eta:[%s]" % (self.eta, ) or "",
                     self.expires and " expires:[%s]" % (self.expires, ) or "")
+
+    def __repr__(self):
+        return '<%s: {name:"%s", id:"%s", args:"%s", kwargs:"%s"}>' % (
+                self.__class__.__name__,
+                self.task_name, self.task_id, self.args, self.kwargs)
+
+    def _get_tracer_args(self, loglevel=None, logfile=None):
+        """Get the :class:`WorkerTaskTrace` tracer for this task."""
+        task_func_kwargs = self.extend_with_default_kwargs(loglevel, logfile)
+        return self.task_name, self.task_id, self.args, task_func_kwargs
+
+    def _set_executed_bit(self):
+        """Set task as executed to make sure it's not executed again."""
+        if self.executed:
+            raise AlreadyExecutedError(
+                   "Task %s[%s] has already been executed" % (
+                       self.task_name, self.task_id))
+        self.executed = True
